@@ -145,6 +145,8 @@
 #include "symtab.h"
 #include "libiberty.h"
 #include "demangle.h"
+#include "elfcpp.h"
+#include "int_encoding.h"
 
 namespace gold
 {
@@ -180,6 +182,11 @@ preprocess_for_unique_sections(const std::vector<Section_id>& id_section,
       section_size_type plen;
       if (section_contents == NULL)
         {
+          // Lock the object so we can read from it.  This is only called
+          // single-threaded from queue_middle_tasks, so it is OK to lock.
+          // Unfortunately we have no way to pass in a Task token.
+          const Task* dummy_task = reinterpret_cast<const Task*>(-1);
+          Task_lock_obj<Object> tl(dummy_task, secn.first);
           const unsigned char* contents;
           contents = secn.first->section_contents(secn.second,
                                                   &plen,
@@ -235,6 +242,11 @@ get_section_contents(bool first_iteration,
 
   if (first_iteration)
     {
+      // Lock the object so we can read from it.  This is only called
+      // single-threaded from queue_middle_tasks, so it is OK to lock.
+      // Unfortunately we have no way to pass in a Task token.
+      const Task* dummy_task = reinterpret_cast<const Task*>(-1);
+      Task_lock_obj<Object> tl(dummy_task, secn.first);
       contents = secn.first->section_contents(secn.second,
                                               &plen,
                                               false);
@@ -263,18 +275,25 @@ get_section_contents(bool first_iteration,
     {
       Icf::Sections_reachable_info v =
         (it_reloc_info_list->second).section_info;
+      // Stores the information of the symbol pointed to by the reloc.
       Icf::Symbol_info s = (it_reloc_info_list->second).symbol_info;
+      // Stores the addend and the symbol value.
       Icf::Addend_info a = (it_reloc_info_list->second).addend_info;
+      // Stores the offset of the reloc.
       Icf::Offset_info o = (it_reloc_info_list->second).offset_info;
+      Icf::Reloc_addend_size_info reloc_addend_size_info =
+        (it_reloc_info_list->second).reloc_addend_size_info;
       Icf::Sections_reachable_info::iterator it_v = v.begin();
       Icf::Symbol_info::iterator it_s = s.begin();
       Icf::Addend_info::iterator it_a = a.begin();
       Icf::Offset_info::iterator it_o = o.begin();
+      Icf::Reloc_addend_size_info::iterator it_addend_size =
+        reloc_addend_size_info.begin();
 
-      for (; it_v != v.end(); ++it_v, ++it_s, ++it_a, ++it_o)
+      for (; it_v != v.end(); ++it_v, ++it_s, ++it_a, ++it_o, ++it_addend_size)
         {
           // ADDEND_STR stores the symbol value and addend and offset,
-          // each atmost 16 hex digits long.  it_a points to a pair
+          // each at most 16 hex digits long.  it_a points to a pair
           // where first is the symbol value and second is the
           // addend.
           char addend_str[50];
@@ -285,6 +304,24 @@ get_section_contents(bool first_iteration,
                    static_cast<long long>((*it_a).first),
 		   static_cast<long long>((*it_a).second),
 		   static_cast<unsigned long long>(*it_o));
+
+	  // If the symbol pointed to by the reloc is not in an ordinary
+	  // section or if the symbol type is not FROM_OBJECT, then the
+	  // object is NULL.
+	  if (it_v->first == NULL)
+            {
+	      if (first_iteration)
+                {
+		  // If the symbol name is available, use it.
+                  if ((*it_s) != NULL)
+                      buffer.append((*it_s)->name());
+                  // Append the addend.
+                  buffer.append(addend_str);
+                  buffer.append("@");
+		}
+	      continue;
+	    }
+
           Section_id reloc_secn(it_v->first, it_v->second);
 
           // If this reloc turns back and points to the same section,
@@ -304,7 +341,12 @@ get_section_contents(bool first_iteration,
             symtab->icf()->section_to_int_map();
           Icf::Uniq_secn_id_map::iterator section_id_map_it =
             section_id_map.find(reloc_secn);
-          if (section_id_map_it != section_id_map.end())
+          bool is_sym_preemptible = (*it_s != NULL
+				     && !(*it_s)->is_from_dynobj()
+				     && !(*it_s)->is_undefined()
+				     && (*it_s)->is_preemptible());
+          if (!is_sym_preemptible
+              && section_id_map_it != section_id_map.end())
             {
               // This is a reloc to a section that might be folded.
               if (num_tracked_relocs)
@@ -331,14 +373,68 @@ get_section_contents(bool first_iteration,
               if (!first_iteration)
                 continue;
 
+              // Lock the object so we can read from it.  This is only called
+              // single-threaded from queue_middle_tasks, so it is OK to lock.
+              // Unfortunately we have no way to pass in a Task token.
+              const Task* dummy_task = reinterpret_cast<const Task*>(-1);
+              Task_lock_obj<Object> tl(dummy_task, it_v->first);
+
               uint64_t secn_flags = (it_v->first)->section_flags(it_v->second);
               // This reloc points to a merge section.  Hash the
               // contents of this section.
-              if ((secn_flags & elfcpp::SHF_MERGE) != 0)
+              if ((secn_flags & elfcpp::SHF_MERGE) != 0
+		  && parameters->target().can_icf_inline_merge_sections ())
                 {
                   uint64_t entsize =
                     (it_v->first)->section_entsize(it_v->second);
-                  long long offset = it_a->first + it_a->second;
+		  long long offset = it_a->first;
+
+                  unsigned long long addend = it_a->second;
+                  // Ignoring the addend when it is a negative value.  See the 
+                  // comments in Merged_symbol_value::Value in object.h.
+                  if (addend < 0xffffff00)
+                    offset = offset + addend;
+
+		  // For SHT_REL relocation sections, the addend is stored in the
+		  // text section at the relocation offset.
+		  uint64_t reloc_addend_value = 0;
+                  const unsigned char* reloc_addend_ptr =
+		    contents + static_cast<unsigned long long>(*it_o);
+		  switch(*it_addend_size)
+		    {
+		      case 0:
+		        {
+                          break;
+                        }
+                      case 1:
+                        {
+                          reloc_addend_value =
+                            read_from_pointer<8>(reloc_addend_ptr);
+			  break;
+                        }
+                      case 2:
+                        {
+                          reloc_addend_value =
+                            read_from_pointer<16>(reloc_addend_ptr);
+			  break;
+                        }
+                      case 4:
+                        {
+                          reloc_addend_value =
+                            read_from_pointer<32>(reloc_addend_ptr);
+			  break;
+                        }
+                      case 8:
+                        {
+                          reloc_addend_value =
+                            read_from_pointer<64>(reloc_addend_ptr);
+			  break;
+                        }
+		      default:
+		        gold_unreachable();
+		    }
+		  offset = offset + reloc_addend_value;
+
                   section_size_type secn_len;
                   const unsigned char* str_contents =
                   (it_v->first)->section_contents(it_v->second,
@@ -394,8 +490,7 @@ get_section_contents(bool first_iteration,
               else if ((*it_s) != NULL)
                 {
                   // If symbol name is available use that.
-                  const char *sym_name = (*it_s)->name();
-                  buffer.append(sym_name);
+                  buffer.append((*it_s)->name());
                   // Append the addend.
                   buffer.append(addend_str);
                   buffer.append("@");
@@ -455,7 +550,7 @@ get_section_contents(bool first_iteration,
 // KEPT_SECTION_ID    : Vector which maps folded sections to kept sections.
 // ID_SECTION         : Vector mapping a section to an unique integer.
 // IS_SECN_OR_GROUP_UNIQUE : To check if a section or a group of identical
-//                            sectionsis already known to be unique.
+//                            sections is already known to be unique.
 // SECTION_CONTENTS   : Store the section's text and relocs to non-ICF
 //                      sections.
 
@@ -603,6 +698,12 @@ Icf::find_identical_sections(const Input_objects* input_objects,
        p != input_objects->relobj_end();
        ++p)
     {
+      // Lock the object so we can read from it.  This is only called
+      // single-threaded from queue_middle_tasks, so it is OK to lock.
+      // Unfortunately we have no way to pass in a Task token.
+      const Task* dummy_task = reinterpret_cast<const Task*>(-1);
+      Task_lock_obj<Object> tl(dummy_task, *p);
+
       for (unsigned int i = 0;i < (*p)->shnum(); ++i)
         {
 	  const char* section_name = (*p)->section_name(i).c_str();
